/* filter_pca.c
 * 
 * Copyright (C) 2015 Tristan Dagobert, Ecole Normale Superieure de Cachan.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @mainpage
 *
 * README.txt:
 * @verbinclude README.txt
 */

/**
 * @file filter_pca.c
 * @brief
 *
 * 
 *
 * @todo 
 *
 * @author Tristan Dagobert <tristan.dagobert@cmla.ens-cachan.fr>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cblas.h>

#include "data_types.h"
#include "io_auxiliary.h"
#include "pca_method.h"
#include "io_parameters.h"
#include "filter_pca.h"

//double compute_mse(float * im, float * im 2
/**
 * @brief main function
 */
int
main (int argc, char **argv)
{
	parameters *param = NULL;

	/* load the command line parameters */
	param = load_parameters (argc, argv);
	verify_parameters (param);

	if (param->algorithm == FIRST_COMPONENT_ALGO)
    {
		/* run the algorithm of Li et al. */
		algorithm_first_component (param);
    }
	else if (param->algorithm == SELECTED_COMPONENT_ALGO)
    {
		/* run the algorithm of SPCA */
		algorithm_selected_component (param);
    }
      
	exit (EXIT_SUCCESS);
}

/* 
 * @brief Decomposition of image upon the basis computed from the principal
 * components of the windowed frames covariance matrix
 * @brief Based on the algorithm described in the following article :
 * "Atmospheric Turbulence Degraded Image Restoration using Principal Components
 * Analysis" by Dalong Li, Russell M. Mersereau and Steven Simske, 2006, GRSL.
 *
 * @param command line parameters
 */

void
algorithm_selected_component (parameters * param)
{
	window_API *win_API = NULL;
	files_list *flist_head, *flist, *flist_prev = NULL;

	float *mu = NULL;

	float *laplacian_mu = NULL;

	float *A = NULL;
	float *eigen_vecs = NULL;
	float *basis_vecs = NULL;

	float *J_im = NULL;

	float * laplacian = NULL;
	float * spmax=NULL, *sp= NULL;    
	float * selected_w = NULL;

	char generic_name[256];
	unsigned int k = 0, m = 0;

	/* load the filenames */
	flist_head = create_files_list (param->list_file_name);
	flist = flist_head;

	/* load the M images of the list */
	for (m = 0; m < param->window_width && flist != NULL; m++)
    {
		printf ("load image %s\n", flist->file_name);
		win_API = load_one_image (win_API, param->window_width, flist->file_name);
		flist_prev = flist;
		flist = flist->next;
    }
	if (flist == NULL && win_API->nimages != param->window_width)
    {
		printf ("Too few images regarding the --n-images-max =%d value\n",
				param->window_width);
		exit (EXIT_FAILURE);
    }
	printf ("%d images of size %d x %d loaded\n", win_API->nimages,
			win_API->nrows, win_API->ncols);

	k = 1;
	flist = flist_prev;
 
	/* laplacian filter */
	laplacian = create_laplacian_filter();

	do
    {
		printf ("#  ======== iteration %d\n", k);

		/* compute mean image i.e. mu */
		mu = compute_mean_image (win_API);

		/* create matrix A */
		A = create_matrix_A (win_API, mu);

		/* compute the eigen vectors v(i) of At.A */
		eigen_vecs = compute_eigens (A, win_API->npixs, win_API->nimages);

		/* compute the orthonormal basis W of A.At */
		basis_vecs = compute_basis (A, win_API->npixs, win_API->nimages,
									eigen_vecs);

		laplacian_mu = compute_laplacian_mu(mu, laplacian, win_API);
		/* compute norm of laplacian mu */
		float norme_lap_mu = 0.0;
		norme_lap_mu = cblas_sdot (win_API->npixs, laplacian_mu, 1, laplacian_mu, 1);
		norme_lap_mu = sqrtf(norme_lap_mu);
		image_scalar_op(&laplacian_mu, laplacian_mu, "/", norme_lap_mu, win_API->npixs);


		/* compute the scalar products of the two first components w1 and w2
		   with Dmu
		*/
		image_image_op(&spmax, laplacian_mu, "$", basis_vecs, win_API->npixs);
		/* copy of w1 */
		image_scalar_op(&selected_w, basis_vecs, "*", 1.0, win_API->npixs);

		/* search for the second component */
		unsigned int i = 1;
		image_image_op(&sp, laplacian_mu, "$", basis_vecs + i * win_API->npixs, 
					   win_API->npixs);

		if (fabsf(*sp) > fabsf(*spmax))
		{
			image_scalar_op(&selected_w, basis_vecs + i * win_API->npixs, "*", 1.0, win_API->npixs);
			*spmax = *sp;
		}
	

		/* computation of J(k) = mu(k) - epsilon.w */
		image_scalar_op(&selected_w, selected_w, "*", param->epsilon, win_API->npixs);
		image_image_op(&J_im, mu, "-", selected_w, win_API->npixs);
 
		sprintf (generic_name, "%s_J_%03d.png", param->output_file_name, k);
		write_image (generic_name, J_im, win_API->ncols, win_API->nrows, PERMIT_SATURATION);

#if defined(EXPLICIT_MODE)

		sprintf (generic_name, "%s_J_%03d.tif", param->output_file_name, k);
		write_image (generic_name, J_im, win_API->ncols, win_API->nrows, NO_REQUANTIZATION);
 
		printf ("statistics_image_J\t");
		print_statistics (J_im, win_API->npixs);

		/* computation of L(k) = mu(k) - epsilon. Lap(mu) / ||Lap(mu)|| */
		image_scalar_op(&laplacian_mu, laplacian_mu, "*", param->epsilon, win_API->npixs);
		image_image_op(&J_im, mu, "-", laplacian_mu, win_API->npixs);
     
		/* image mu corrected by its laplacian */
		sprintf (generic_name, "%s_L_%03d.png", param->output_file_name, k);
		write_image (generic_name, J_im, win_API->ncols, win_API->nrows, PERMIT_SATURATION);

		sprintf (generic_name, "%s_L_%03d.tif", param->output_file_name, k);
		write_image (generic_name, J_im, win_API->ncols, win_API->nrows, NO_REQUANTIZATION);

		printf ("statistics_image_L\t");
		print_statistics (J_im, win_API->npixs);
      
		/* the well oriented selected component w */
		sprintf (generic_name, "%s_wa_%03d.png", param->output_file_name, k);
		write_image (generic_name, selected_w, win_API->ncols,
				     win_API->nrows, WITH_REQUANTIZATION);
		printf ("statistics_image_wa\t");
		print_statistics (selected_w, win_API->npixs);
      
		/* the raw component w1 */
		sprintf (generic_name, "%s_w1_%03d.png", param->output_file_name, k);
		write_image (generic_name, basis_vecs, win_API->ncols,
				     win_API->nrows, WITH_REQUANTIZATION);
		printf ("statistics_image_w1\t");
		print_statistics (basis_vecs, win_API->npixs);

		/* the raw component w2 */
		sprintf (generic_name, "%s_w2_%03d.png", param->output_file_name, k);
		write_image (generic_name, basis_vecs+win_API->npixs, win_API->ncols,
				     win_API->nrows, WITH_REQUANTIZATION);
		printf ("statistics_image_w2\t");
		print_statistics (basis_vecs+win_API->npixs, win_API->npixs);
      
		/* the mean image = mu(k) */
		sprintf (generic_name, "%s_mu_%03d.png", param->output_file_name, k);
		write_image (generic_name, mu, win_API->ncols,
				     win_API->nrows, PERMIT_SATURATION);
		printf ("statistics_image_mu\t");
		print_statistics (mu, win_API->npixs);

		/* the laplacian of mu */
		sprintf (generic_name, "%s_lmu_%03d.png", param->output_file_name, k);
		write_image (generic_name, laplacian_mu, win_API->ncols,
				     win_API->nrows, WITH_REQUANTIZATION);
		printf ("statistics_image_laplacian\t");
		print_statistics (laplacian_mu, win_API->npixs);
      
#endif

		/* clean memory */
		free(mu);
		free(laplacian_mu);
		free (A);
		free (eigen_vecs);
		free (basis_vecs);

		/* load the next image into the sliding window and 
		 * remove the head of the window, according to
		 * a FIFO mechanism
		 */
		flist = flist->next;
		if (flist != NULL)
		{
			win_API =
				load_one_image (win_API, param->window_width, flist->file_name);
		}

		k++;
    }
	while (flist != NULL);

	free(sp);
	free(spmax);
	free (J_im);    
	free(laplacian);
	free(selected_w);
	free_files_list(flist_head);
	free_window_API (win_API);
	free_parameters (param);
	return;
}

/* 
 * @brief Based on the algorithm described in the following article :
 * "Atmospheric Turbulence Degraded Image Restoration using Principal Components
 * Analysis" by Dalong Li, Russell M. Mersereau and Steven Simske, 2006, GRSL.
 *
 * @param command line parameters
 */

void
algorithm_first_component (parameters * param)
{
	window_API *win_API = NULL;
	files_list *flist_head, *flist, *flist_prev = NULL;

	float *mean_im = NULL;

	float *A = NULL;
	float *eigen_vecs = NULL;
	float *basis_vecs = NULL;

	float *J_im = NULL;

	char generic_name[256];
	unsigned int k = 0, m = 0;

	/* load the filenames */
	flist_head = create_files_list (param->list_file_name);
	flist = flist_head;

	/* load the M first images into the sliding window */
	for (m = 0; m < param->window_width && flist != NULL; m++)
    {
		printf ("load image %s\n", flist->file_name);
		win_API = load_one_image (win_API, param->window_width, flist->file_name);
		flist_prev = flist;
		flist = flist->next;
    }
	if (flist == NULL && win_API->nimages != param->window_width)
    {
		printf ("Too few images regarding the --n-images-max =%d value\n",
				param->window_width);
		exit (EXIT_FAILURE);
    }
	printf ("%d images of size %d x %d loaded\n", win_API->nimages,
			win_API->nrows, win_API->ncols);

	k = 1;
	flist = flist_prev;
	do
    {
		printf ("#  ======== iteration %d\n", k);

		/* compute mean image i.e. mu(k) */
		mean_im = compute_mean_image (win_API);

		/* create matrix A */
		A = create_matrix_A (win_API, mean_im);

		/* compute the eigen vectors of At.A */
		eigen_vecs = compute_eigens (A, win_API->npixs, win_API->nimages);

		/* compute the image J(k) = A(k).v(1) + mu(k) */
		J_im =
			compute_image_from_v1 (A, 1, win_API->npixs, win_API->nimages,
								   eigen_vecs, mean_im);
		sprintf (generic_name, "%s_J_%03d.png", param->output_file_name, k);
		write_image (generic_name, J_im, win_API->ncols, win_API->nrows, PERMIT_SATURATION);
#if defined(EXPLICIT_MODE)
		printf ("statistics_image_J\t");
		print_statistics (J_im, win_API->npixs);
#endif
		free (J_im);

		/* clean memory */
		free (A);
		free (eigen_vecs);
		free (basis_vecs);

		/* load the next image into the sliding window and 
		 * remove the head of the window, according to
		 * a FIFO mechanism
		 */
		flist = flist->next;
		if (flist != NULL)
		{
			win_API =
				load_one_image (win_API, param->window_width, flist->file_name);
		}

		k++;
    }
	while (flist != NULL);

	free_window_API (win_API);
	free_parameters (param);
	return;
}
