/* io_auxiliary.c
 * 
 * Copyright (C) 2015 Tristan Dagobert, Ecole Normale Superieure de Cachan.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @mainpage
 *
 * README.txt:
 * @verbinclude README.txt
 */

/**
 * @file io_auxiliary.c
 * @brief
 *
 * 
 *
 * @todo 
 *
 * @author Tristan Dagobert <tristan.dagobert@cmla.ens-cachan.fr>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "data_types.h"
#include "iio.h"
#include "io_auxiliary.h"

/**
 * @brief print a matrix
 * 
 * @param desc matrix name
 * @param nrows
 * @param ncols
 * @param mat the matrix
 * @param lda number of data cells  between two columns
 */
void
print_matrix (char *desc, int nrows, int ncols, float *mat, int lda)
{
  int i, j;
  printf ("\n %s\n", desc);
  for (i = 0; i < nrows; i++)
    {
      printf ("%02d\t", i);
      for (j = 0; j < ncols; j++)
	printf (" %10e", mat[i + j * lda]);
      printf ("\n");
    }
}

/** 
 * @brief free the parameters structure
 */
void
free_parameters (parameters * param)
{
  /* not need to free param->list_file_name and
     param->output_file_name because these two variables
     are directly linked to ARGV
   */
  free (param);
}

/** 
 * @brief free the window structure
 */
void
free_window (window * win)
{
  free (win->im);
  free (win);
  return;
}

/** 
 * @brief free the window_API structure
 * @param win_API
 */
void
free_window_API (window_API * win_API)
{
  window *win_ptr = NULL, *win_head = NULL;

  win_head = win_API->win;
  while (win_head != NULL)
    {
      win_ptr = win_head->next;
      free_window (win_head);
      win_head = win_ptr;
    }
  free (win_API);
  return;
}

/**
 * @brief Create the list of images filenames according to the type
 * of temporal extension of the sequence.
 * @param list_file_name file containing the path of the images
 * @return the trigger of the images filenames
 */
files_list *
create_files_list (char *list_file_name)
{
  FILE *file_ptr = NULL;
  files_list *flist_elem = NULL;
  files_list *flist_head = NULL;
  files_list *flist_end = NULL;

  char png_file_name[256];
  char *file_name = NULL;
  int l = 0;
  int is_first_time = 1;

  file_ptr = fopen (list_file_name, "r");
  if (file_ptr == NULL)
    {
      printf ("list file not found !!\n");
      exit (EXIT_FAILURE);
    }

  /* read the file containing the names of the images */
  while (!feof (file_ptr))
    {
      if (fscanf (file_ptr, "%s", png_file_name) == EOF)
	{
	  continue;
	}

      /* new element allocation */
      flist_elem = (files_list *) malloc (sizeof (files_list));
      if (flist_elem == NULL)
	{
	  printf ("allocation failed !!\n");
	  exit (EXIT_FAILURE);
	}

      /* new element initialization */
      l = strlen (png_file_name);
      file_name = (char *) malloc ((l + 1) * sizeof (char));
      if (file_name == NULL)
	{
	  printf ("allocation failed !!\n");
	  exit (EXIT_FAILURE);
	}
      strcpy (file_name, png_file_name);
      flist_elem->file_name = file_name;
      flist_elem->next = NULL;
      flist_elem->prev = NULL;

      /* the first time : the classic pointers juggling */
      if (is_first_time == 1)
	{
	  flist_head = flist_elem;
	  flist_end = flist_elem;
	  is_first_time = 0;
	}
      else
	{
	  flist_end->next = flist_elem;
	  flist_elem->prev = flist_end;
	  flist_end = flist_end->next;
	}
    }
  fclose (file_ptr);

  return flist_head;
}


/**
 * @brief load the image into the gliding window according
 * to a FIFO mechanism. If the window is full, the oldest image will 
 * be remove and the new image will be added at the end.
 * @param w_API the gliding window API
 * @param nimages_max size of the window i.e the maximal number of images the window 
 *        have to contain
 * @param png_file_name file name of the image to load into the window
 * @return the new gliding window
 */
window_API *
load_one_image (window_API * w_API, unsigned int nimages_max,
		char *png_file_name)
{
  window_API *win_API = NULL;
  window *win_elem = NULL;
  window *win_ptr = NULL;

  float *im = NULL;
  unsigned int nrows = 0, ncols = 0, npixs = 0;
  int nr, nc;

  win_API = w_API;

  /* read the new image */
  im = iio_read_image_float (png_file_name, &nc, &nr);
  //im = read_png_f32_gray (png_file_name, &nc, &nr);
  if (im == NULL)
    {
      printf ("allocation failed for reading image %s\n", png_file_name);
      exit (EXIT_FAILURE);
    }
  else
    {
      printf ("allocation succeded for image %s\n", png_file_name);
      fflush(stdout);
    }

  /* rescaling from [a;b] (supposed to be in [0;255]) to [a/255;b/255] */
  dynamic_rescaling (im, nc, nr, 1, 1.0/255.0);
  nrows = nr;
  ncols = nc;
  npixs = nrows * ncols;

  /* images must have same dimensions */
  if (win_API != NULL)
    {
      if (win_API->nrows != nrows || win_API->ncols != ncols)
	{
	  printf ("Are you joking ? Images must have the same dimensions !!\n");
	  exit (EXIT_FAILURE);
	}
    }
  else
    {
      /* if the window is not still definite */
      win_API = (window_API *) malloc (sizeof (window_API));
      if (win_API == NULL)
	{
	  printf ("allocation failed !!\n");
	  exit (EXIT_FAILURE);
	}
      win_API->nrows = nrows;
      win_API->ncols = ncols;
      win_API->npixs = npixs;
      win_API->nimages = 0;
      win_API->win = NULL;
      win_API->win_end = NULL;
    }

  /* structure allocation */
  win_elem = (window *) malloc (sizeof (window));
  if (win_elem == NULL)
    {
      printf ("allocation failed !!\n");
      exit (EXIT_FAILURE);
    }
  /* fields initialization */
  win_elem->next = NULL;
  win_elem->im = im;

  /* add the new element at the end of the list */
  if (win_API->win == NULL)
    {
      win_API->win = win_elem;
    }
  else
    {
      win_API->win_end->next = win_elem;
    }
  win_API->win_end = win_elem;
  win_API->nimages++;

  /* remove the first element if needed */
  if (win_API->nimages > nimages_max)
    {
      win_ptr = win_API->win;
      win_API->win = win_API->win->next;
      free_window (win_ptr);
      win_API->nimages--;
    }
  return win_API;
}

/**
 * @brief Requantize the image dynamic between 0 and 255
 * @param image the image
 * @param width image width
 * @param height image height
 * @param channel_number number of image channels
 */
void
dynamic_requantization (float *image, int width, int height, int channel_number)
{
  int i = 0;

  /* Computing the mininum and max of image */
  float min = image[0];
  float max = image[0];

  for (i = 1; i < width * height * channel_number; i++)
    {
      if (image[i] < min)
	min = image[i];
      if (image[i] > max)
	max = image[i];
    }

  /* Actually changing image values */
  for (i = 0; i < width * height * channel_number; i++)
    {
      image[i] = (255 * (image[i] - min) / (max - min));
    }
  return;
}

/**
 * @brief Quantify the image dynamic which is originally between [dmin, dmax]
 * to [dmin/255, dmax/255]
 * @param image the image
 * @param width image width
 * @param height image height
 * @param channel_number number of image channels
 */
void
dynamic_rescaling (float *image, int width, int height, int channel_number, float scale_factor)
{
  int i = 0;
  for (i = 0; i < width * height * channel_number; i++)
    {
      image[i] *= scale_factor;
    }
  return;
}

void
dynamic_saturation (float *image, int width, int height, int channel_number, 
		    float minimum, float maximum)
{
  int i = 0;
  for (i = 0; i < width * height * channel_number; i++)
    {
      if (image[i] < minimum)
	{
	  image[i] = minimum;
	}
      else if (maximum < image[i])
	{
	  image[i] = maximum;
	}
    }
  return;
}

/**
 * @brief Write an image to the PNG format
 * @param generic_name name of the image file
 * @param im the image to be written
 * @param ncols image width
 * @param nrows image length
 * @param dynamic_transformation
 * @return nothing
 */
void
write_image(char *generic_name, float *im, int ncols, int nrows, int dynamic_transformation)
{
  float *im_tmp = NULL;
  //int info;

  /* make an image copy */
  im_tmp = (float *) malloc (ncols * nrows * sizeof (float));
  if (im_tmp == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }
  memcpy (im_tmp, im, ncols * nrows * sizeof (float));
 
  /* rescaling from [a;b] (which normally in [0; 1]) to [a*255;b*255] */
  dynamic_rescaling (im_tmp, ncols, nrows, 1, 255.0);

  /* dynamic transformation */
  if (dynamic_transformation == WITH_REQUANTIZATION)
    {
      dynamic_requantization (im_tmp, ncols, nrows, 1);
    }
  else if (dynamic_transformation == PERMIT_SATURATION)
    {
      dynamic_saturation(im_tmp, ncols, nrows, 1, 0.0, 255.0);
    }

  /* writing */
  iio_save_image_float_split (generic_name, im_tmp, ncols, nrows, 1);
//  info = write_png_f32 (generic_name, im_tmp, ncols, nrows, 1);
//  if (info != 0)
//    {
//      printf ("writing problem !!\n");
//      exit (EXIT_FAILURE);
//    }
//
  free (im_tmp);
  return;
}

void free_files_list(files_list *flist_head)
{
  files_list *flist = NULL;
  files_list *flist_ptr = NULL;
  
  flist = flist_ptr = flist_head;

  while(flist_head->next != NULL)
    {
      while(flist->next != NULL)
        {
          flist_ptr = flist;
          flist=flist->next;
        }
      /* suppression du dernier élément */
      free(flist->file_name);
      free(flist);
      flist_ptr->next = NULL;

      /* on recommence depuis le début */
      flist = flist_ptr = flist_head;
    }
  free(flist_head->file_name);
  free(flist_head);
}
