/* io_parameters.c
 * 
 * Copyright (C) 2015 Tristan Dagobert, Ecole Normale Superieure de Cachan.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @mainpage
 *
 * README.txt:
 * @verbinclude README.txt
 */

/**
 * @file io_parameters.c
 * @brief
 *
 * 
 *
 * @todo 
 *
 * @author Tristan Dagobert <tristan.dagobert@cmla.ens-cachan.fr>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "io_parameters.h"

#define VERSION "1.0"
/**
 * @brief display the error command line
 * @param argv command line parameters
 * @return
 */
void
display_error (char **argv)
{
  printf ("%s : option unknown\n", argv[0]);
  display_usage (argv);
  printf ("To know more, type : %s --help\n", argv[0]);
  return;
}

/**
 * @brief display the program version
 * @param argv command line parameters
 * @return
 */
void
display_usage (char **argv)
{
  printf ("Usage: %s --version | --help | ", argv[0]);
  printf ("--list FILE ");
  printf ("--output PREFIX ");
  printf ("--window-width SIZE ");
  printf ("--epsilon EPSILON ");
  printf ("\n");
  printf ("| ");
  printf ("--list FILE ");
  printf ("--output PREFIX ");
  printf ("--window-width SIZE\n");
  return;
}

/**
 * @brief display the program version
 * @return
 */
void
display_version (void)
{

  printf ("IPOL filter_pca.exe %s\n", VERSION);
  printf
    ("Copyright (c) 2015 Tristan Dagobert, CMLA, Ecole Normale Suprieure de Cachan\n\n");
  printf ("This program is distributed in the hope that it will be useful,\n");
  printf ("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  printf ("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n");
  printf ("GNU General Public License for more details.\n");
  return;
}

/**
 * @brief display the help 
 * @param argv command line parameters
 * @return
 */
void
display_help (char **argv)
{
  display_usage (argv);
  printf ("\n");
  printf("This program is associated with the algorithm described in the ");
  printf("publication entitled \n << Study of the Principal Component ");
  printf("Analysis Method for the Correction of Images\nDegraded by Turbulences >>.\n");
  printf("It creates an image J with the prefix PREFIX from an image stack size SIZE,\n");
  printf("whose names are contained in the FILE file. Parameter EPSILON is used for\n");
  printf("the SPCA method, while without Li's method is applied. \n");
  printf("For more details about the meaning of these algorithms, please \n");
  printf("read the associated article.\n");
  printf("NOTE : the input and output images MUST BE be png images. \n");

  printf ("Example :\n%s --list images.list --output prefix_images ", argv[0]);
  printf ("--window-width 5 --epsilon 0.5;\n\n");

  printf ("Options :\n");
  printf ("--version            display the program version\n");
  printf ("--help               display this help\n");
  printf ("\n");
  printf ("Report bugs to: tristan.dagobert@cmla.ens-cachan.fr\n");
  printf ("Home page: <http://www.ipol.im/>\n");
  return;
}

/**
 * @brief process the command line
 * @param argc number of parameters
 * @param command line parameters
 * @return the parameters
 */
parameters *
load_parameters (int argc, char **argv)
{
  parameters *param = NULL;

  /* processing depending on the parameters number */
  switch (argc)
    {
    case 1:
      {
	display_error (argv);
	exit (EXIT_SUCCESS);
	break;
      }
    case 2:
      {
	if (strcmp ("--help", argv[1]) == 0)
	  {
	    display_help (argv);
	    exit (EXIT_SUCCESS);
	  }
	else if (strcmp ("--version", argv[1]) == 0)
	  {
	    display_version ();
	    exit (EXIT_SUCCESS);
	  }
	else
	  {
	    display_error (argv);
	    exit (EXIT_SUCCESS);
	  }
	break;
      }
      /* Li' s method */
    case 7:
      {
	if (strcmp ("--list", argv[1]) != 0
	    || strcmp ("--output", argv[3]) != 0
	    || strcmp ("--window-width", argv[5]) != 0)
	  {
	    display_error (argv);
	    exit (EXIT_FAILURE);
	  }
	else
	  {
	    /* load parameters */
	    param = (parameters *) malloc (sizeof (parameters));
	    if (param == NULL)
	      {
		printf ("allocation failed !!\n");
		exit (EXIT_FAILURE);
	      }
	    param->list_file_name = argv[2];
	    param->output_file_name = argv[4];
	    param->window_width = atoi (argv[6]);
	    param->algorithm = FIRST_COMPONENT_ALGO;
	  }
	break;
      }
      /* SPCA method */
    case 9:
      {
	if (strcmp ("--list", argv[1]) != 0
	    || strcmp ("--output", argv[3]) != 0
	    || strcmp ("--window-width", argv[5]) != 0
	    || strcmp ("--epsilon", argv[7]) != 0)
	  {
	    display_error (argv);
	    exit (EXIT_FAILURE);
	  }
	else
	  {
	    /* load parameters */
	    param = (parameters *) malloc (sizeof (parameters));
	    if (param == NULL)
	      {
		printf ("allocation failed !!\n");
		exit (EXIT_FAILURE);
	      }
	    param->list_file_name = argv[2];
	    param->output_file_name = argv[4];
	    param->window_width = atoi (argv[6]);
	    param->epsilon = atof (argv[8]);
	    param->algorithm = SELECTED_COMPONENT_ALGO;
	  }
	break;
      }
    default:
      {
	display_error (argv);
	exit (EXIT_SUCCESS);
      }
    }
  return param;
}

/**
 * @brief Verify the values coherence of command line parameters.
 * Program is stopped in case of incoherence.
 * @param param command line parameters
 * @return
 */
void
verify_parameters (parameters * param)
{
  if (param->window_width < 3)
    {
      printf ("Parameter SIZE must be greater than 2.\n");
      exit (EXIT_FAILURE);
    }

  return;
}
